from binascii import unhexlify
from cgi import parse_header
import datetime
import json
import logging
import platform


_allowed_methods = ("POST", "PUT", "OPTIONS")
_datetime_format = "%Y-%m-%dT%H:%M:%S%z"


def _parse_datetime(datestr):
    if not datestr:
        return None
    # tweak timezone indicator so that %z handles it
    if datestr.endswith("Z"):  # Z -> +0000
        datestr = datestr[:-1] + "+0000"
    elif datestr[-3] == ":" and datestr[-6] in ("+", "-"):  # +xx:yy -> +xxyy
        datestr = datestr[0:-3] + datestr[-2:]
    return datetime.datetime.strptime(datestr, _datetime_format)

def _read_nbytes(fileobj, size):
    data = b""
    while size > 0:
        tmp = fileobj.read(size)
        if not tmp:
            break
        size = size - len(tmp)
        data = data + tmp
    return data

class Receiver(object):

    log = logging.getLogger(__name__)
    _server_header = ("Server", "Quriiri-RX-Python/1.0.1 %s/%s" % (
        platform.python_implementation(), platform.python_version()))
    _base_headers = [_server_header, ("Content-Length", "0")]

    def handle_input(self, data):
        """
        Handle parsed input data. This implementation does nothing but
        logs what it receives, override in subclasses as appropriate.
        """
        self.log.info("RX: %s" % data)

    def parse_json(self, data, charset):
        """
        Parse input from the given byte data using
        the given charset. Raises and logs an exception on error.
        """
        try:
            data = data.decode(charset)
            data = json.loads(data)
        except:
            self.log.exception("JSON parse error")
            raise
        for k, v in data.items():
            try:
                if k.endswith("time"):
                    data[k] = _parse_datetime(v)
                elif k == "udh":
                    data[k] = unhexlify(v)
            except:
                self.log.exception("Parameter conversion error for %s" % k)
                raise
        return data

    def wsgi_application(self, environ, start_response):
        """
        WSGI entry point.
        """
        if environ.get("REQUEST_METHOD") not in _allowed_methods:
            start_response("405 Method Not Allowed", [
                ("Allow", ", ".join(_allowed_methods)),
            ] + self._base_headers)
            return []
        if environ["REQUEST_METHOD"] == "OPTIONS":
            start_response("200 OK", [
                ("Allow", ", ".join(_allowed_methods)),
            ] + self._base_headers)
            return []

        content_type, params = parse_header(environ.get("CONTENT_TYPE", ""))
        if content_type.lower() != "application/json":
            start_response("415 Unsupported Media Type", self._base_headers)
            return []

        content_length = environ.get("CONTENT_LENGTH")
        if content_length is None:
            start_response("411 Length Required", self._base_headers)
            return []

        try:
            data = self.parse_json(
                _read_nbytes(environ.get("wsgi.input"), int(content_length)),
                params.get("charset", "UTF-8")
            )
        except:
            start_response("400 Bad Request", [
                ("Warning", '199 - "JSON parsing error"'),
            ] + self._base_headers)
            return []

        try:
            self.handle_input(data)
        except:
            self.log.exception("Input handling error")
            start_response("500 Internal Server Error", [
                ("Warning", '199 - "Input handling error"'),
            ] + self._base_headers)
            return []

        start_response("204 No Content", [self._server_header])
        return []
